<{{ __quikui_component_name__|lower }}
{%- if quikui_css_classes %} class="{{ quikui_css_classes }}"{% endif %}
{%- if quikui_extra_attributes %} {{ quikui_extra_attributes }}{% endif %}>
  {%- for item in items -%}{{ item }}{%- endfor -%}
</{{ __quikui_component_name__|lower }}>
