{% extends 'AppBase.html' %} {% block header %}
<title>Live-stream Dashboard</title>
{% endblock header %} {% block content %}
<div id="my-chart" style="width: 100%; max-width: 300px; margin: 0 auto;">
  <table class="charts-css line show-labels">
    <thead>
      <tr>
        <th scope="col">Time</th>
        <th scope="col">Data</th>
      </tr>
    </thead>
    <tbody
      hx-ext="sse"
      sse-connect="/stream-data"
      sse-swap="message"
      hx-swap="beforeend"
    >
    {% for item in data %} {{ item }} {% endfor %}
    </tbody>
  </table>
</div>
{% endblock content %}
