{% extends "FormInput.html" %}
{% block input %}
<select name="{{ name }}" {% if required %}required{% endif %} {{ __extra_attrs__ }}>
{% for option in options -%}
  <option
    value="{{ option.value }}"
    {% if option.value == selected %}selected{% endif %}
    {% if option.value in disabled %}disabled{% endif %}>
    {{ option.label }}
  </option>
{%- endfor %}
</select>
{% endblock %}
