{% if label -%}
  {{ label }}
  {% if add_break -%}<br>{%- endif %}
{%- endif %}
{% block input %}
<input
  {{ __extra_attrs__ }}
  type="{{ type }}"
  name="{{ name }}"
  {% if value %}value="{{ value }}"{% endif %}
  {% if required %}required{% endif %}>
{% endblock %}
